/*
 tel210.c

 Author: Allan Latham <alatham@flexsys-group.com>

*/ 

#define SCREENLOG "/tmp/tel210.log1"
#define KEYBLOG "/tmp/tel210.log2"
#define KEYSENDLOG "/tmp/tel210.log3"

#define FALSE 0
#define TRUE 1
#define MAX_HOST 64
#define BUFFSIZE 128
#define MAX_LINES 24
#define MAX_COLUMNS 80

#include "version.h"
#include "config.h"
#include "d410names.h"

#include <curses.h>
#include <string.h>
#include <stdlib.h> 
#include <stdio.h>   
#include <unistd.h>   
#include <stddef.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/termios.h>

/* prototypes */

void display_arguments();
void display_hello();
void forward_traffic();
void get_arguments(int, char**);
void get_pty();
void set_noecho(int);
void die(int);
void init_screen();
void restore_screen();
void clear_screen();
void rack_up();
void erase_line();
void set_raw(int);
unsigned char readpty();
unsigned char readtty();   
unsigned char readtty0();   
int which_window();
void reset_windows();
void refresh_windows();
void set_attributes();
void write_win_addr();
void read_win_addr();
void cursor_up();
void cursor_down();
void cursor_left();
void cursor_right();
void under_on();
void under_off();
void dim_on();
void dim_off();
void newline();
void creturn();
void erase_window();
void erase_unprotect();
void window_home();
void roll_enable();
void roll_disable();
void read_model();              
void reverse_on();           
void reverse_off();         
void scroll_up();             
void scroll_down();          
void insert_char();         
void delete_char();        
void line_1();                
int  backspace(char*);
void scommand();
void command_106();        
void command_107();         
void fill_screen();        
void command_10677();    
void d200_reset();      
void set_windows();  
void erase_screen();       
void screen_home();       
void insert_line();      
void delete_line();   
void change_attr();         
void read_hso();           
void write_scr_addr();   
void set_cur_type();     
void define_char();        
void sel_char_set();        
void set_scr_rate();        
void sel_7_8_bit();         
void set_margins();      
void set_amargins();     
void insert_line();      
void delete_line();     
void show_cols();       
void read_scr_addr();   
void read_char_rem();   
void reserve_char();    
void set_kb_lang();   
void bar_1();               
void command_10776();     
void command_10777();   
void read_cur_attr();  
void cursor_track();     
void command_107160(); 
void read_cur_loc();     
void write_cur_loc();       
void set_pattern();
void check_scroll();
          
/* globals */

WINDOW *wptr;

char *fk  = "Xqrstuvwxyz{|}~pXXXYZabcdefghijklmn`XXXXX123456789:;<=>0XXXXX!\"#$%&'()*+,-. XXXXX";

//               0   1   2   3   4   5   6   7   8   9
int fcnt[90] = { 0,  0,  0,  0,  0,  0,  0,  6,  7,  8, \
                 9, 10,  0, 11, 12, 13, 14,  0, 15,  0, \
                 0,  0,  0, 19, 20, 21, 22, 23, 24, 25, \
                26, 27, 28, 29, 30, 31, 32, 33, 34, 35, \
                 0,  0,  0,  0,  0, 41, 42, 43, 44, 45, \
                46, 47, 48, 49, 50, 51, 52, 53, 54, 55, \
                 0,  0,  0,  0,  0, 61, 62, 63, 64, 65, \
                66, 67, 68, 69, 70, 71, 72, 73, 74, 75, \
                 0,  0,  0,  0,  0,  0,  0,  0,  0,  0 };  

int wstart_line [MAX_LINES];
int wno_of_lines [MAX_LINES];
int wrow [MAX_LINES];
int wcol [MAX_LINES];
WINDOW *wsptr [MAX_LINES];

int kill;
int lines_left;
int cwind;
int cwr;
int cwr0;
int cwr16;
int dchar;
int fcn;
int col;
int row;
int attr;
int roll_enabled;
int scroll_enabled;

FILE *lf1;
FILE *lf2;
FILE *lf3;

unsigned char single[2];

int curh;
int curv;

unsigned char curh_1, curh_2, curv_1, curv_2;

char *av0;
char **saved_argv;

FILE *pmfp;

int pmfd;
int psfd;
int screen_on;

int command_state;

/* arguments to main */

int debug = 0;
unsigned int port = 0;
char host[MAX_HOST];

/* ---------------------------------------------------------------- */
int main(int ac, char **av)
{
 char sport[6];

/* FIXME - Signal handlers */

 kill = 0;
 display_hello();
 get_arguments(ac, av);
 display_arguments();
 get_pty();

 screen_on = FALSE;

 if (fork())
 {
  close(psfd);
  forward_traffic();  
  die(-1);
 }
 else
 {
  close(pmfd);
  close(0);
  close(1);
  close(2);
  dup(psfd);
  dup(psfd);
  dup(psfd);

  set_noecho(0);

  if (port)
  {
   sprintf(sport, "%d", port);
   execl("/usr/bin/telnet", "/usr/bin/telnet -8E", host, sport, NULL);
  }
  else
  {
   execl("/usr/bin/telnet", "/usr/bin/telnet -8E", host, NULL);
  }
  fprintf(stderr, "exec: %.100s\n", strerror(errno));
  
  die(-1);
 }
 die(0);
 exit(0);
}

/* ---------------------------------------------------------------- */
void forward_traffic()
{
 fd_set rfds;
 struct timeval tv;
 unsigned char buf[BUFFSIZE];
 int i, max, len, ret, ok, funkey;

 d200_reset();

 if (debug > 4)
 {
  lf1 = fopen (SCREENLOG, "w");
  lf2 = fopen (KEYBLOG, "w");
  lf3 = fopen (KEYSENDLOG, "w");
 }

 max = pmfd + 1;

 for (;;)
 {
  FD_ZERO(&rfds);
  FD_SET(pmfd, &rfds);
  FD_SET(0, &rfds);
  tv.tv_sec = 5;
  tv.tv_usec = 0;
  ret = select(max, &rfds, NULL, NULL, &tv);
// no timeout version  ret = select(max, &rfds, NULL, NULL, NULL);
  if (FD_ISSET(pmfd, &rfds))
  {
   single[0] = readpty();
   switch (command_state)
   {
    case 0:     /* command_state */
    {
     if (single[0] == BACKSPACE)
     {
      ok = backspace(single);
     }
     else
     {
      if (single[0] <= HIGHEST_CONTROL)
      {
       ok = FALSE;
       switch(single[0])
       {
        case NULLCHAR:                             break;
        case PRINT_FORM:                           break;
        case BLINK_ENABLE:                         break;
        case BLINK_DISABLE:                        break;
        case READ_WIN_ADDR:   read_win_addr();     break;
        case BELL:                                 break; 
        case WINDOW_HOME:     window_home();       break;
        case NEWLINE:         newline();           break;
        case ERASELINE:       erase_line();        break;
        case ERASE_WINDOW:    erase_window();      break;
        case RETURN:          creturn();           break;
        case BLINK_ON:                             break;
        case BLINK_OFF:                            break;
        case WRITE_WIN_ADDR:  write_win_addr();    break;
        case PRINT_WINDOW:                         break;
        case ROLL_ENABLE:     roll_enable();       break;
        case ROLL_DISABLE:    roll_disable();      break;
        case UNDER_ON:        under_on();          break;
        case UNDER_OFF:       under_off();         break;
        case CURSOR_UP:       cursor_up();         break;
        case CURSOR_RIGHT:    cursor_right();      break;
        case CURSOR_LEFT:     cursor_left();       break;
        case CURSOR_DOWN:     cursor_down();       break;
        case HIGH_ON:         dim_on();            break;
        case HIGH_OFF:        dim_off();           break;
        case COMMAND:         scommand();          break;
        default:
        {          
         if (debug > 3)
         {
          fprintf(stderr,"0x%x", single[0]);
         }
         ok = TRUE;
         single[0] = '~';
        }
       }
      }
      else
      {
       ok = TRUE;
      }
     }
     break;
    }
    case 1:     /* command_state  COMMAND */
    {
     ok = FALSE;
     command_state = 0;
     switch(single[0])
     {
      case READ_MODEL:      read_model();        break;      
      case REVERSE_ON:      reverse_on();        break;     
      case REVERSE_OFF:     reverse_off();       break;   
      case COMMAND_106:     command_106();       break;     
      case COMMAND_107:     command_107();       break;    
      case SCROLL_UP:       scroll_up();         break;       
      case SCROLL_DOWN:     scroll_down();       break;     
      case INSERT_CHAR:     insert_char();       break;    
      case DELETE_CHAR:     delete_char();       break;   
      case LINE_1:          line_1();            break;       
      case SHIFT_OUT:                            break;       
      case SHIFT_IN:                             break;          
      default: 
      {
       ok = TRUE;
      }
     }
     break;
    }
    case 2:     /* command_state COMMAND 106 */
    {
     ok = FALSE;
     command_state = 0;
     switch(single[0])
     {
      case UART_TEST:                            break;
      case DISPLAY_SW:                           break;
      case FILL_SCREEN:    fill_screen();        break; 
      case COMMAND_10677:  command_10677();      break;
      case ANSI_MODE:                            break; 
      case RESET:          d200_reset();         break; 
      case SET_WINDOWS:    set_windows();        break; 
      case SCROLL_LEFT:                          break; 
      case SCROLL_RIGHT:                         break; 
      case ERASE_SCREEN:   erase_screen();       break; 
      case ERASE_UNPROT:   erase_unprotect();    break; 
      case SCREEN_HOME:    screen_home();        break; 
      case INSERT_LINE:    insert_line();        break; 
      case DELETE_LINE:    delete_line();        break; 
      case NORM_SPACE:                           break; 
      case COMP_SPACE:                           break; 
      case PROTECT_ON:                           break;  
      case PROTECT_OFF:                          break; 
      case CHANGE_ATTR:    change_attr();        break; 
      case READ_HSO:       read_hso();           break; 
      case WRITE_SCR_ADDR: write_scr_addr();     break; 
      case SET_CUR_TYPE:   set_cur_type();       break; 
      case DEFINE_CHAR:    define_char();        break; 
      case SEL_CHAR_SET:   sel_char_set();       break; 
      case SET_SCR_RATE:   set_scr_rate();       break; 
      case SEL_7_8_BIT:    sel_7_8_bit();        break; 
      case PROTECT_EN:                           break; 
      case PROTECT_DIS:                          break; 
      case SET_MARGINS:    set_margins();        break; 
      case SET_AMARGINS:   set_amargins();       break; 
      case RES_MARGINS:                          break; 
      case INS_LINE_M:     insert_line();        break; 
      case DEL_LINE_M:     delete_line();        break; 
      case HS_DISABLE:                           break; 
      case HS_ENABLE:                            break; 
      case SHOW_COLS:      show_cols();          break; 
      case PPT_ON_1:                             break; 
      case PPT_OFF_1:                            break; 
      case READ_SCR_ADDR:  read_scr_addr();      break; 
      case READ_CHAR_REM:  read_char_rem();      break; 
      case RESERVE_CHAR:   reserve_char();       break; 
      case SET_KB_LANG:    set_kb_lang();        break; 
      default:
      {
       ok = TRUE;
      }
     }
     break;
    }
    case 3:     /* command_state  COMMAND 107 */
    {
     ok = FALSE;
     command_state = 0;
     switch(single[0])
     {
      case BAR:            bar_1();              break; 
      case LINE_2:         line_1();             break;
      case COMMAND_10776:  command_10776();      break;
      case COMMAND_10777:  command_10777();      break;
      case READ_CUR_ATTR:  read_cur_attr();      break;
      case CURSOR_RESET:                         break;
      case CURSOR_ON:                            break;
      case CURSOR_OFF:                           break;
      case CURSOR_TRACK:   cursor_track();       break;
      case COMMAND_107160: command_107160();     break;
      default:
      {
       ok = TRUE;
      }
     }
     break;
    }
    case 4:     /* command_state COMMAND 106 77 */
    {
     ok = FALSE;
     command_state = 0;
     switch(single[0])
     {
      case PPT_OFF_2:                            break; 
      case PPT_ON_2:                             break;
      case WIN_BIT_DUMP:                         break;
      case FORM_BIT_DUMP:                        break;
      default:
      {
       ok = TRUE;
      }
     }
     break;
    }
    case 5:     /* command_state COMMAND 107 76 */
    {
     ok = FALSE;
     command_state = 0;
     switch(single[0])
     {
      case WRITE_CUR_LOC: write_cur_loc();       break;  
      default:
      {
       ok = TRUE;
      }
     }
     break;
    }
    case 6:     /* command_state COMMAND 107 77 */
    {
     ok = FALSE;
     command_state = 0;
     switch(single[0])
     {
      case READ_CUR_LOC: read_cur_loc();         break;  
      default:
      {
       ok = TRUE;
      }
     }
     break;
    }
    case 7:     /* command_state COMMAND 107 160 */
    {
     ok = FALSE;
     command_state = 0;
     switch(single[0])
     {
      case SET_PATTERN: set_pattern();           break;  
      default:
      {
       ok = TRUE;
      }
     }
     break;
    }
    case 21:    /* command_state */
    case 22:    /* command_state */
    {
     write_win_addr();
     break;
    }
    case 25:    /* command_state */
    case 26:    /* command_state */
    case 27:    /* command_state */
    {
     line_1();
     break;
    }
    case 31:    /* command_state */
    case 32:    /* command_state */
    case 33:    /* command_state */
    {
     set_windows();  
     break;
    }
    case 38:    /* command_state */
    {
     set_cur_type();  
     break;
    }
    case 41:    /* command_state */
    case 42:    /* command_state */
    case 43:    /* command_state */
    case 44:    /* command_state */
    case 45:    /* command_state */
    {
     change_attr();
     break;
    }
    case 46:    /* command_state */
    {
     sel_7_8_bit();        
     break;
    }
    case 47:    /* command_state */
    {
     set_scr_rate();        
     break;
    }
    case 49:    /* command_state */
    {
     define_char();
     break;
    }
    case 50:    /* command_state */
    {
     set_kb_lang();
     break;
    }
    case 51:    /* command_state */
    case 52:    /* command_state */
    case 53:    /* command_state */
    case 54:    /* command_state */
    {
     write_scr_addr();
     break;
    }
    case 58:    /* command_state */
    {
     cursor_track(); 
     break;
    }
    case 60:    /* command_state */
    {
     bar_1();
     break;
    }
    case 61:    /* command_state */
    case 62:    /* command_state */
    case 63:    /* command_state */
    case 64:    /* command_state */
    {
     set_margins();
     break;
    }
    case 68:    /* command_state */
    {
     sel_char_set();
     break;
    }
    case 69:    /* command_state */
    {
     sel_char_set();
     break;
    }
    case 71:    /* command_state */
    case 72:    /* command_state */
    case 73:    /* command_state */
    case 74:    /* command_state */
    case 75:    /* command_state */
    case 76:    /* command_state */
    {
     set_amargins();
     break;
    }
    case 81:    /* command_state */
    case 82:    /* command_state */
    case 83:    /* command_state */
    case 84:    /* command_state */
    {
     show_cols();
     break;
    }
   }
   if (ok)
   {
    single[1] = 0;
    if (col < MAX_COLUMNS)
    {
     set_attributes();
     mvaddstr (row, col, single);
     col++;                         
     if (col == MAX_COLUMNS)
     {
      check_scroll();
     }
    }                               
    else                            
    {
     row++;
     col = 0;
     if (row >= MAX_LINES)               
     {                                   
      rack_up();                         
      set_attributes();
      mvaddstr (row, col, single);
      col++;
     }
    }
   }
   else
   {
    move (row, col);
   }
//   clearok (stdscr,TRUE);
   refresh_windows();
   refresh ();
  }

  if (FD_ISSET(0, &rfds))
  {
   buf[0] = readtty();
   len = 1;
   funkey = 0;
   switch (buf[0])
   {
    case NEWLINE:
     buf[0] = RETURN;
     break;
    case RETURN:
     buf[0] = NEWLINE;
     break;
    case ESCAPE:
     buf[1] = readtty0();
     if (buf[1] == 0)
     {
      break;
     }
     len = 2;
     if (buf[1] == '[')
     {
      buf[2] = readtty();
      len = 3;
      switch (buf[2])
      {
       case '[':
        buf[3] = readtty();
        len = 4;
        switch (buf[3])
        { 
         case 'A':
          funkey = 1;
          break;
         case 'B':
          funkey = 2;
          break;
         case 'C':
          funkey = 3;
          break;
         case 'D':
          funkey = 4;
          break;
         case 'E':
          funkey = 5;
          break;
        }
        break;
       case '1':
       case '2':
       case '3':
       case '4':
       case '5':
       case '6':
       case '7':
       case '8':
       case '9':
        buf[3] = readtty();
        len = 4;
        switch (buf[3])
        { 
         case '1':
         case '2':
         case '3':
         case '4':
         case '5':
         case '6':
         case '7':
         case '8':
         case '9':
         case '0':
          fcn = ((buf[2] - '1') * 10) + (buf[3] - '0');
          funkey = fcnt [fcn];
          break;
        }
        buf[4] = readtty();
        len = 5;
        if (buf[4] != '~')
        {
         funkey = 0;
        }
        break;
       case 'A':
        buf[0] = CURSOR_UP;
        len = 1;
        break;
       case 'B':
        buf[0] = CURSOR_DOWN;
        len = 1;
        break;
       case 'C':
        buf[0] = CURSOR_RIGHT;
        len = 1;
        break;
       case 'D':
        buf[0] = CURSOR_LEFT;
        len = 1;
        break;
      }
     }
    break;
   }
   switch (funkey)
   {
    case 0:
     kill = 0;
     break;
    default:
     len = 2;
     buf[0] = 0x1e;
     buf[1] = fk[funkey];
     if (buf[1] == 'X')
     {
      len = 0;
     }
     if (buf[1] == 'Y')
     {
      len = 0;

/* diagnostics with alt-F9 */

      fprintf(stderr, "col = %d, row = %d, win = %d, state = %d\n", 
              col, row, which_window(), command_state);
     }
     if (buf[1] == 'Z')
     {
      len = 0;
      if (kill) die(0);
      kill = 1;

/* kill with alt-F10 */

      fprintf(stderr,
         "Press alt-F10 again to terminate, any other key to continue.\n");
     }
     break;
   }
   if (len)
   {
    if (debug > 4)
    {
     for(i = 0; i < len; i++) 
     {
      fputc (buf[i], lf3);
     }
    } 
    if (write(pmfd, buf, len) != len)
    {
     if (debug > 1)
     {
      fprintf(stderr, "write(pmfd): %.100s\n", strerror(errno));
     }
     die (1);
    }
   }
  }
 }
}

/* ---------------------------------------------------------------- */
void get_pty()
{
 char *ptys = "0123456789abcdef";
 int i, c;
 static char buf[16];

 for (c = 'p'; c <= 's'; c++)
 { 
  for (i = 0; i < 16; i++)
  {
   sprintf(buf, "/dev/pty%c%c", c, ptys[i]);
   if ((pmfd = open(buf, O_RDWR)) >= 0)
   {
    sprintf(buf, "/dev/tty%c%c", c, ptys[i]);
    if ((psfd = open(buf, O_RDWR | O_NOCTTY | O_NDELAY)) < 0)
    {
     close(pmfd);
    } else {
     pmfp = fdopen(pmfd, "r+");
     return;
    }
   }
  }
 }
}

/* ---------------------------------------------------------------- */
void get_arguments(int ac, char **av)
{
 extern char *optarg;
 extern int optind;

 int opt;

 saved_argv = av;

 if (strchr(av[0], '/'))
 {
  av0 = strrchr(av[0], '/') + 1;
 }
 else
 {
  av0 = av[0];
 }

 while ((opt = getopt(ac, av, "d:p:h?")) != EOF)
 {
  switch (opt)
  {
   case 'p':
    port = atoi(optarg);
    break;
   case 'd':
    debug = atoi(optarg);
    break;
   case '?':
   case 'h':
   default:
    fprintf(stderr, "tel210 ");
    fprintf(stderr, "Usage: %s [options] host\n", av0);
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -d debug_level  2 or above = keep in foreground\n");
    fprintf(stderr, "  -h or -?        this help message\n");
    fprintf(stderr, "  -p port         client mode port\n");
    fprintf(stderr, "Returns:\n");
    fprintf(stderr, "   0 normal exit\n");
    fprintf(stderr, "  -1 if there was an error\n\n");
    exit(-1);
  }
 }

 if (optind < ac)
 {
  strncpy(host, av[optind], MAX_HOST);
 }
 else
 {
  fprintf(stderr, "host missing\n");
  die(-1);
 }

 if (host[MAX_HOST - 1] != 0)
 {
  fprintf(stderr, "host too long\n");
  die(-1);
 }

 optind++;

 if (optind < ac)
 {
  fprintf(stderr, "Extra argument %s.\n", av[optind]);
  die(-1);
 }

}

/* ---------------------------------------------------------------- */
void display_hello()
{
 printf("\nTEL210 %s, Copyright (C) 1997,8 Allan Latham <alatham@flexsys-group.com>\n\n", VERSION);
 printf("      This program is a telnet which behaves like a DG Dasher D210\n");
 printf("      TEL210 comes with ABSOLUTELY NO WARRANTY. Use it at YOUR OWN RISK.\n");
 printf("      This is free software; you are welcome to redistribute it\n");
 printf("      or modify it under certain conditions; see source for details.\n\n");
 printf("      Some of the above words are protected in law as trademarks.\n");
 printf("      This product has in no way been approved, acknowledged or endorsed\n");
 printf("      by any of the trademark holders, nor indeed by anyone at all.\n\n");
}

/* ---------------------------------------------------------------- */
void display_arguments()
{
 if (port)
 {
  printf("\nConnecting to %s port %d\n", host, port);
 }
 else
 {
  printf("\nConnecting to %s \n", host);
 }
 sleep(3);
}

/* ---------------------------------------------------------------- */
void set_noecho(int fd)
{
 struct termios stermios;

 if (tcgetattr(fd, &stermios) < 0)
 {
  fprintf(stderr, "get attr: %.100s\n", strerror(errno));
  die(-1);
 }
 stermios.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
 stermios.c_oflag &= ~(ONLCR);
 if (tcsetattr(fd, TCSANOW, &stermios) < 0)
 {
  fprintf(stderr, "set attr: %.100s\n", strerror(errno));
  die(-1);
 }
}
 
/* ---------------------------------------------------------------- */
void set_raw(int fd)
{
 struct termios stermios;

 if (tcgetattr(fd, &stermios) < 0)
 {
  fprintf(stderr, "get raw: %.100s\n", strerror(errno));
  die(-1);
 }
 stermios.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
 stermios.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
 stermios.c_cflag &= ~(CSIZE | PARENB);
 stermios.c_cflag |= CS8;

// stermios.c_oflag &= ~(OPOST);
// stermios.c_oflag &= ~(ONLCR);

 stermios.c_cc[VMIN] = 0;
// stermios.c_cc[VMIN] = 1;
 stermios.c_cc[VTIME] = 2;
// ajl stermios.c_cc[VTIME] = 0;
 if (tcsetattr(fd, TCSAFLUSH, &stermios) < 0)
 {
  fprintf(stderr, "set raw: %.100s\n", strerror(errno));
  die(-1);
 }
}
 
/* ---------------------------------------------------------------- */
void die(int reply)
{
 if (screen_on)
 {
  restore_screen();
 }
 printf("\nEnd of TEL210 - Normal terminal mode resumed\n\n");
 exit(reply);
}

/* ---------------------------------------------------------------- */
void init_screen()
{
 wptr = initscr ();
 start_color();
 init_pair(1,COLOR_GREEN, COLOR_BLACK);
 init_pair(2,COLOR_YELLOW, COLOR_BLACK);
 attr = 0;
 set_attributes();
// nonl ();
// raw ();
// noecho ();
 set_raw (0);
 wclear (stdscr);
 screen_on = TRUE;
 scrollok(stdscr,TRUE);
 clear_screen();
}

/* ---------------------------------------------------------------- */
void d200_reset()
{
 reset_windows();
 col = row = command_state = 0;
 scroll_enabled = roll_enabled = TRUE;
 init_screen();
}

/* ---------------------------------------------------------------- */
void restore_screen()
{
 endwin ();
 screen_on = FALSE;
}

/* ---------------------------------------------------------------- */
void clear_screen()
{
 col = 0;
 row = 0;
 attr = 0;
 attrset (attr);
 erase ();
}

/* ---------------------------------------------------------------- */
void rack_up()
{
 scroll (stdscr);
 row = MAX_LINES-1;
 col = 0;
 clrtoeol();
}

/* ---------------------------------------------------------------- */
void erase_line()
{
 clrtoeol();
}

/* ---------------------------------------------------------------- */
unsigned char readpty()
{
 unsigned char c;

 if ((read(pmfd, &c, 1)) < 1)
 {
  die (0);
 }
 if (debug > 4)
 {
  fputc (c, lf1);
 } 
 return (c);
}  

/* ---------------------------------------------------------------- */
unsigned char readtty()
{
 unsigned char c;
 int reply = 0;

 while (reply != 1)
 {
  reply = read(0, &c, 1);
  if (reply < 0)
  {
   if (debug > 1)
   {
    fprintf(stderr, "read(0): %.100s\n", strerror(errno));
   }
   die (1);
  }
 }

 if (debug > 4)
 {
  fputc (c, lf2);
 } 

 return(c);
}

/* ---------------------------------------------------------------- */
unsigned char readtty0()
{
 unsigned char c;
 int reply = 0;

 reply = read(0, &c, 1);
 if (reply < 0)
 {
  if (debug > 1)
  {
   fprintf(stderr, "read0(0): %.100s\n", strerror(errno));
  }
  die (1);
 }

 if (reply == 1)
 { 
  if (debug > 4)
  {
   fputc (c, lf2);
  }
 }
 else
 {
  c = 0;
 } 

 return(c);
}

/* ---------------------------------------------------------------- */
void set_attributes()
{
/* xxxxxxxx
          1 = reverse
         1  = bold
        1   = underline
*/
 switch (attr)
 {
  case 0:
   attrset(COLOR_PAIR(1));
   attron (A_BOLD);
   break;
  case 1:
   attrset(COLOR_PAIR(1));
   attron (A_REVERSE);
   attron (A_BOLD);
   break;
  case 2:
   attrset(COLOR_PAIR(1));
   break;
  case 3:
   attrset(COLOR_PAIR(1));
   attron (A_REVERSE);
   break;
  case 4:
   attrset(COLOR_PAIR(2));
   attron (A_BOLD);
   break;
  case 5:
   attrset(COLOR_PAIR(2));
   attron (A_BOLD);
   attron (A_REVERSE);
   break;
  case 6:
   attrset(COLOR_PAIR(2));
   break;
  case 7:
   attrset(COLOR_PAIR(2));
   attron (A_REVERSE);
   break;
 }
}

/* ---------------------------------------------------------------- */
void write_win_addr()
{
 int w, lns;

 switch (command_state)
 {
  case 0:
  {
   command_state = 21;
   break;
  }
  case 21:
  {
   curh = single[0];
   command_state = 22;
   break;
  }
  case 22:
  {
   curv = single[0];
   w = which_window();
   lns = wno_of_lines[w] + 1;
   if ((curh < MAX_COLUMNS) && (curv < lns))
   {
    col = curh;
    row = curv + wstart_line[w];
   }
   command_state = 0;
   break;
  }
 }
}

/* ---------------------------------------------------------------- */
void read_win_addr()
{
 int w;
 unsigned char scolrow[3];

 w = which_window();
 
 scolrow[0] = 31;
 scolrow[1] = col;
 scolrow[2] = row - wstart_line[w];

 if (write(pmfd, scolrow, 3) != 3)
 {
  if (debug > 1)
  {
   fprintf(stderr, "write(pmfd): %.100s\n", strerror(errno));
  }
  die (1);
 }
}

/* ---------------------------------------------------------------- */
void cursor_left()
{
 int w, wr;

 w = which_window();
 wr = row - wstart_line[w];

 col--;
 if (col < 0)
 {
  wr--;
  col = MAX_COLUMNS-1;
  if (wr < 0)
  {
   row = wstart_line[w] + wno_of_lines[w] - 1;
  }
  else
  {
   row--;
  }
 }
}

/* ---------------------------------------------------------------- */
void cursor_right()
{
 int w, wr, mrow;

 w = which_window();
 wr = row - wstart_line[w];

 col++;
 if (col >= MAX_COLUMNS)
 {
  col = 0;
  wr++;
  mrow = wno_of_lines[w] - 1;
  if (wr > mrow)
  {
   row = wstart_line[w];
  }
  else
  {
   row++;
  }
 }
}

/* ---------------------------------------------------------------- */
void cursor_up()
{
 int w, wr;

 w = which_window();
 wr = row - wstart_line[w];

 wr--;
 if (wr < 0)
 {
  row = wstart_line[w] + wno_of_lines[w] - 1;
 }
 else
 {
  row--;
 }
}

/* ---------------------------------------------------------------- */
void cursor_down()
{   
 int w, wr, mrow;

 w = which_window();
 wr = row - wstart_line[w];
 mrow = wno_of_lines[w] - 1; 

 wr++;
 if (wr > mrow)
 {
  row = wstart_line[w];
 }
 else
 {
  row++;
 }
}

/* ---------------------------------------------------------------- */
void under_on()
{
 attr = attr | 0x04;
}

/* ---------------------------------------------------------------- */
void under_off()
{
 attr = attr & ~0x04;
}

/* ---------------------------------------------------------------- */
void dim_on()
{
 attr = attr | 0x02;
}

/* ---------------------------------------------------------------- */
void dim_off()
{
 attr = attr & ~0x02;
}

/* ---------------------------------------------------------------- */
void newline()
{
 int w, wr, mrow;

 w = which_window();
 wr = row - wstart_line[w];
 mrow = wno_of_lines[w] - 1;

 col = 0;
 wr++;
 if (wr > mrow)
 {
  scroll_up();    
 }
 else
 {
  row++;
 }
}

/* ---------------------------------------------------------------- */
void creturn()
{
 col = 0;
}

/* ---------------------------------------------------------------- */
void window_home()
{
 int w;

 w = which_window();

 col = 0;
 row = wstart_line[w];
}

/* ---------------------------------------------------------------- */
void erase_unprotect()
{
 int w, wmax;

 w = which_window();
 wmax = wstart_line[w] + wno_of_lines[w];

 col = 0;

 for ( ; row < wmax; row++)
 {
  move (row, col);
  clrtoeol();
 }
 row = wstart_line[w];
}

/* ---------------------------------------------------------------- */
void erase_window()
{
 int w, wmax;

 w = which_window();
 wmax = wstart_line[w] + wno_of_lines[w];

 col = 0;
 for (row = wstart_line[w]; row < wmax; row++)
 {
  move (row, col);
  clrtoeol();
 }
 row = wstart_line[w];
}

/* ---------------------------------------------------------------- */
int backspace(char*ch)
{
 if (col == 0)
 {
  return (FALSE);
 }
 else
 {
  col--;
  ch[0] = ' ';
  return (TRUE);
 }
}

/* ---------------------------------------------------------------- */
void roll_enable()
{
 roll_enabled = TRUE;
}

/* ---------------------------------------------------------------- */
void roll_disable()
{
 roll_enabled = FALSE;
}

/* ---------------------------------------------------------------- */
void scommand()
{
 command_state = 1;
}

/* ---------------------------------------------------------------- */
void read_model()              
{
 unsigned char model[6];
 
 model[0] = 036;
 model[1] = 0157;
 model[2] = 043;
 model[3] = 052;
 model[4] = 0x51;
 model[5] = 0x5a;

 if (write(pmfd, model, 6) != 6)
 {
  if (debug > 1)
  {
   fprintf(stderr, "write(pmfd): %.100s\n", strerror(errno));
  }
  die (1);
 }
}

/* ---------------------------------------------------------------- */
void reverse_on()           
{
 attr = attr | 0x01;
}

/* ---------------------------------------------------------------- */
void reverse_off()         
{
 attr = attr & ~0x01;
}

/* ---------------------------------------------------------------- */
void scroll_up()             
{
 int w;

 w = which_window();
 if (wsptr[w])
 {
  wscrl (wsptr[w],1);
 }
 else
 {
  scrl (1);
 }
}

/* ---------------------------------------------------------------- */
void scroll_down()          
{
 int w;

 w = which_window();
 if (wsptr[w])
 {
  wscrl (wsptr[w],-1);
 }
 else
 {
  scrl (-1);
 }
}

/* ---------------------------------------------------------------- */
void insert_char()         
{
 insch(' ');
}

/* ---------------------------------------------------------------- */
void delete_char()        
{
 delch();
}

/* ---------------------------------------------------------------- */
void line_1()                
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 25;
   break;
  }
  case 25:
  case 26:
  case 27:
  {
   if (single[0] == 0)
   {
    if (command_state == 27)
    {
     command_state = 0;
    }
    else
    {
     command_state++;
    }
   }
   else
   {
    command_state = 25;
   }
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void command_106()        
{
 command_state = 2;
}

/* ---------------------------------------------------------------- */
void command_107()         
{
 command_state = 3;
}

/* ---------------------------------------------------------------- */
void fill_screen()        
{
 int i;
 char fill [MAX_COLUMNS+1];

 switch (command_state)
 {
  case 0:
  {
   command_state = 24;
   break;
  }
  case 24:
  {
   for (i = 0; i < MAX_COLUMNS; i++)
   {
    fill[i] = single[0];
   }
   fill[MAX_COLUMNS] = 0;
   for (i = 0; i < MAX_LINES; i++)
   {
    mvaddstr (i, 0, fill);
   }
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void command_10677()    
{
 command_state = 4;
}

/* ---------------------------------------------------------------- */
int which_window()
{
 int i, j;

 for (i = 0; i < MAX_LINES; i++) 
 {
  j = wstart_line[i] + wno_of_lines[i];
  if (row < j)
  {
   return (i);
  }
 }
 return(0);
}

/* ---------------------------------------------------------------- */
void refresh_windows()
{
 int i;

 for (i = 0; i < MAX_LINES; i++) 
 {
  if (wsptr[i])
  {
   wrefresh(wsptr[i]);
  }
 }
}

/* ---------------------------------------------------------------- */
void reset_windows()
{
 int i;

 for (i = 0; i < MAX_LINES; i++) 
 {
  wstart_line[i] = 0;
  wno_of_lines[i] = 0;
  wrow[i] = 0;
  wcol[i] = 0;
  wsptr[i] = NULL;
 }
 wno_of_lines[0] = MAX_LINES;
}

/* ---------------------------------------------------------------- */
void set_windows()  
{
 int i;

 switch (command_state)
 {
  case 0:
  {
   for (i = 0; i < MAX_LINES; i++) 
   {
    if (wsptr[i])
    {
     delwin(wsptr[i]);
    }
   }
   reset_windows();
   wno_of_lines[0] = 0;
   lines_left = MAX_LINES;
   cwind = 0;
   command_state = 31;
   break;
  }
  case 31:
  {
   cwr16 = single[0] - '0';
   command_state = 32;
   break;
  }
  case 32:
  {
   cwr0 = single[0] - '0';
   command_state = 33;
   break;
  }
  case 33:
  {
   cwr = cwr0 + (cwr16 * 16);
   if ((cwr > lines_left) || (cwr == 0))
   {
    cwr = lines_left;
   }
   wno_of_lines[cwind] = cwr;
   if (cwind != 0)
   {
    wstart_line[cwind] = wstart_line[cwind - 1] + wno_of_lines[cwind - 1];
   }
   wsptr[cwind] = 
          subwin(wptr,wno_of_lines[cwind],MAX_COLUMNS,wstart_line[cwind],0);
   scrollok (wsptr[cwind], TRUE);
   lines_left = lines_left - cwr;
   if (lines_left == 0)
   {
    command_state = 0;
   }
   else
   {
    cwind++;
    command_state = 31;
   }
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void erase_screen()       
{
 clear_screen();
}

/* ---------------------------------------------------------------- */
void screen_home()       
{
 col = 0;
 row = 0;
}

/* ---------------------------------------------------------------- */
void insert_line()      
{
 int w;

 w = which_window();
 if (wsptr[w])
 {
  winsertln(wsptr[w]);
 }
 else
 {
  insertln();
 }
}

/* ---------------------------------------------------------------- */
void delete_line()   
{
 int w;

 w = which_window();
 if (wsptr[w])
 {
  wdeleteln(wsptr[w]);
 }
 else
 {
  deleteln();
 }
}

/* ---------------------------------------------------------------- */
void change_attr()         
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 41;
   break;
  }
  case 41:
  case 42:
  case 43:
  case 44:
  {
   command_state++;
   break;
  }
  case 45:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void read_hso()           
{
 unsigned char hso[5];
 
 hso[0] = 036;
 hso[1] = 0157;
 hso[2] = 072;
 hso[3] = 0;
 hso[4] = 0;

 if (write(pmfd, hso, 5) != 5)
 {
  if (debug > 1)
  {
   fprintf(stderr, "write(pmfd): %.100s\n", strerror(errno));
  }
  die (1);
 }
}

/* ---------------------------------------------------------------- */
void write_scr_addr()   
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 51;
   break;
  }
  case 51:
  {
   curh_1 = single[0];
   command_state = 52;
   break;
  }
  case 52:
  {
   curh_2 = single[0];
   command_state = 53;
   break;
  }
  case 53:
  {
   curv_1 = single[0];
   command_state = 54;
   break;
  }
  case 54:
  {
   curv_2 = single[0];

   if ((curh_1 == '?') && (curh_2 == '?'))
   {
    curh = col;
   }
   else
   {
    curh_1 = curh_1 - '0';
    curh_2 = curh_2 - '0';
    curh = curh_2 + (curh_1 * 16);
   }

   if ((curv_1 == '?') && (curv_2 == '?'))
   {
    curv = row;
   }
   else
   {
    curv_1 = curv_1 - '0';
    curv_2 = curv_2 - '0';
    curv = curv_2 + (curv_1 * 16);
   }

   if ((curh < MAX_COLUMNS) && (curv < MAX_LINES))
   {
    col = curh;
    row = curv;
   }
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void set_cur_type()     
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 38;
   break;
  }
  case 38:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void define_char()        
{
 switch (command_state)
 {
  case 0:
  {
   dchar = 25;
   command_state = 49;
   break;
  }
  case 49:
  {
   dchar--;
   if (dchar == 0)
   {
    command_state = 0;
   }
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void sel_char_set()        
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 68;
   break;
  }
  case 68:
  {
   command_state = 69;
   break;
  }
  case 69:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void set_scr_rate()        
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 47;
   break;
  }
  case 47:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void sel_7_8_bit()         
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 46;
   break;
  }
  case 46:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void set_margins()      
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 61;
   break;
  }
  case 61:
  case 62:
  case 63:
  {
   command_state++;
   break;
  }
  case 64:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void set_amargins()     
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 71;
   break;
  }
  case 71:
  case 72:
  case 73:
  case 74:
  case 75:
  {
   command_state++;
   break;
  }
  case 76:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void show_cols()       
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 81;
   break;
  }
  case 81:
  case 82:
  case 83:
  {
   command_state++;
   break;
  }
  case 84:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void read_scr_addr()  
{
 unsigned char scolrow[7];
 
 scolrow[0] = 036;
 scolrow[1] = 157;
 scolrow[2] = 043;
 scolrow[3] = col / 10;
 scolrow[4] = col - scolrow[3];
 scolrow[5] = row / 10;
 scolrow[6] = row - scolrow[5];

 scolrow[3] = scolrow[3] + '@';
 scolrow[4] = scolrow[4] + '@';
 scolrow[5] = scolrow[5] + '@';
 scolrow[6] = scolrow[6] + '@';

 if (write(pmfd, scolrow, 7) != 7)
 {
  if (debug > 1)
  {
   fprintf(stderr, "write(pmfd): %.100s\n", strerror(errno));
  }
  die (1);
 }
}

/* ---------------------------------------------------------------- */
void read_char_rem()   
{
 unsigned char scount[5];
 
 scount[0] = 036;
 scount[1] = 157;
 scount[2] = 071;
 scount[3] = '@';
 scount[4] = '@';

 if (write(pmfd, scount, 5) != 5)
 {
  if (debug > 1)
  {
   fprintf(stderr, "write(pmfd): %.100s\n", strerror(errno));
  }
  die (1);
 }
}

/* ---------------------------------------------------------------- */
void reserve_char()    
{
 return;
}

/* ---------------------------------------------------------------- */
void set_kb_lang()    
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 50;
   break;
  }
  case 50:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void bar_1()               
{
 switch (command_state)
 {
  case 0:
  {
   dchar = 13;
   command_state = 60;
   break;
  }
  case 60:
  {
   dchar--;
   if (dchar == 0)
   {
    command_state = 0;
   }
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void command_10776()     
{
 command_state = 5;
}

/* ---------------------------------------------------------------- */
void command_10777()   
{
 command_state = 6;
}

/* ---------------------------------------------------------------- */
void read_cur_attr()  
{
 return;
}

/* ---------------------------------------------------------------- */
void cursor_track()     
{
 switch (command_state)
 {
  case 0:
  {
   command_state = 58;
   break;
  }
  case 58:
  {
   command_state = 0;
   break;
  }
  default:
  {
   command_state = 0;
  }
 }
}

/* ---------------------------------------------------------------- */
void command_107160() 
{
 command_state = 7;
}

/* ---------------------------------------------------------------- */
void read_cur_loc()     
{
 return;
}

/* ---------------------------------------------------------------- */
void write_cur_loc()       
{
 return;
}

/* ---------------------------------------------------------------- */
void set_pattern()
{
 return;
}

/* ---------------------------------------------------------------- */
void check_scroll()
{
 int w, wr, mrow;

 w = which_window();
 wr = row - wstart_line[w];
 mrow = wno_of_lines[w] - 1;

 col = 0;
 wr++;
 if (wr > mrow)
 {
  if (roll_enabled)
  {
   scroll_up();    
  }
  else
  {
   row = wstart_line[w];
  }
 }
 else
 {
  row++;
 }
}
